      subroutine ardrive(N, A,
     1            nev, ncv, lworkl, resid, v, workd, workl, info, ido,
     2            selection, d, ierr, 
     3            maxit)
c* Driver for ARPACK: all calculations are done in FORTRAN
c* memory allocations done in C
c* Arguments
c* N - matrix size
c* A - full NxN matrix, column major format
c* nev, ncv, lworkl, resid, v, workd, workl, info, ido, selection, d, ierr
c* - all the above are usual parameters in ARPACK
c* maxit - max outer iterations of ARPACK

c* ================================================================
c*   Fitscorr: subspace decomposition of FITS files
c*  Copyright (C) 2006-2008 Sarod Yatawatta <sarod@users.sf.net>  
c* $Id: ardrive.f,v 1.3 2005/04/19 00:32:25 sarod Exp $
c* ================================================================

c* Arguments
      integer N,
     1        nev, ncv, lworkl, info, ido, 
     2        maxit
      double precision A(N*N), resid(N), v(N,ncv),
     1  workd(3*N), workl(lworkl), selection(ncv), d(ncv,2)
c* Local variables
      integer ldv, ipntr(11), iparam(11) 
      double precision tol, sigma
      character bmat*1, which*2
      logical rvec

c* debugging -- uncomment the following
c*      include 'debug.h'
c*      ndigit = -3
c*      logfil = 6
c*      msgets = 0
c*      msaitr = 0
c*      msapps = 0
c*      msaupd = 1
c*      msaup2 = 0
c*      mseigt = 0
c*      mseupd = 0

c* Initialize Locals
      bmat ='I'
      which='LM'
      tol=0.0D+0
      ldv=N
c* set globals
      info=0
      ido=0
c* Set Algorithm Mode
      iparam(1)=1
      iparam(3)=maxit
      iparam(7)=1
c* Main loop
   10 continue
      call dsaupd(ido, bmat, N, which, nev, tol, resid,
     1       ncv, v, ldv, iparam, ipntr, workd, workl,
     2       lworkl, info)

      if (ido .eq. -1 .or. ido .eq. 1) then
        call matproduct(N,A,workd(ipntr(1)),workd(ipntr(2)))
        go to 10
      end if

      if (info .eq. -9999) then
        print *,'Not enough memory current size=',iparam(5)
      end if

c* If no errors, calculate eigenvectors
      if (info .eq. 0) then
        rvec =.true.
        call dseupd ( rvec, 'All', selection, d, v, ldv, sigma,
     &         bmat, N, which, nev, tol, resid, ncv, v, ldv,
     &         iparam, ipntr, workd, workl, lworkl, ierr )

      end if
c* end of program
      return
      end


      subroutine matproduct(N, A, x, y)
c*  Calculate matrix product Ax=>y, A is column major NxN
c*  Array indices are given in Fortran format, so runs from 1 to N, unlike
c*  C format.

c* Arguments
      integer N
      double precision A(N*N), x(N), y(N)
c* Local variables
      integer i
c* zero y
      do 10 i=1,N
        y(i)=0.0
   10 continue   
c* calculate product
      do 20 i=1,N
c*  y =y +x(i)*A(:,i)
        call daxpy(N,x(i),A((i-1)*N+1),1,y,1)
   20 continue   

c* end subroutine
      return
      end
