/*  Fitscorr: subspace decomposition of FITS files
 
 Copyright (C) 2006-2008 Sarod Yatawatta <sarod@users.sf.net>  
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 $Id$
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <dirent.h>
#include <unistd.h>

#include <fitsio.h>
#include <fitsio2.h>

#include <stddef.h>
#include <sys/types.h>
#include <sys/wait.h>

#ifndef ABS
#define ABS(x) \
          ((x)>=0? (x):-(x))
#endif

#define SUBTRACT_MODES 0
#define PRESERVE_MODES 1
#define PRESERVE_UNIQUE_MODE  2

#define USE_BLAS 1

//#define DEBUG
/***** LAPACK */
int dgesvd(char JOBU, char JOBVT, int M, int N, double *A, int LDA, double *S,
				 double *U, int LDU, double *VT, int LDVT, double *WORK, int LWORK) {
	int info;
	extern void dgesvd_(char *JOBU, char *JOBVT, int *M, int *N, double *A, int *LDA, double *S,
				 double *U, int *LDU, double *VT, int *LDVT, double *WORK, int *LWORK, int *INFO);

	
	dgesvd_(&JOBU, &JOBVT, &M, &N, A, &LDA, S,
				 U, &LDU, VT, &LDVT, WORK, &LWORK, &info);


	return info;
}


/* divide and conquor */
int dgesdd(char JOBZ, int M, int N, double *A, int LDA, double *S, double *U,
				int LDU, double *VT, int LDVT, double *WORK, int LWORK, int *IWORK) {

				int info;
        extern void dgesdd_(char *JOBZ, int *M, int *N, double *A, int *LDA, double *S, double *U,
				int *LDU, double *VT, int *LDVT, double *WORK, int *LWORK, int *IWORK, int *INFO);

   dgesdd_(&JOBZ, &M, &N, A, &LDA, S, U,
				&LDU, VT, &LDVT, WORK, &LWORK, IWORK, &info);

	 return info;

}

#ifdef USE_BLAS
/* blas */
extern double dnrm2_(int *N, const double *x, int *incx);
extern void daxpy_(int *N, double *alpha, double *x, int *incx, double *y, int *incy);
extern double ddot_(int *N, double *x, int *incx, double *y, int *incy);
extern void dscal_(int *N, double *alpha, double *x, int *incx);
extern void dcopy_(int *N, double *x, int *incx, double *y, int *incy);

/* BLAS wrappers */
/* norm ||x|| */
double
my_norm(int N, double *x) {
  int i=1;
  return dnrm2_(&N,x,&i);
}
/* y = a.x + y */
void
my_daxpy(int N, double *x, double a, double *y) {
  int i=1;
  daxpy_(&N,&a,x,&i,y,&i);
}
/* dot x^T y */
double
my_ddot(int N, double *x, double *y) {
  int i=1;
  return ddot_(&N,x,&i,y,&i);
}
/* x = a. x */
void
my_dscal(int N, double a, double *x) {
  int i=1;
  dscal_(&N,&a,x,&i);
}
/* y = x */
void
my_dcopy(int N, double *x, double *y) {
  int i=1;
  dcopy_(&N,x,&i,y,&i);
}
#endif /* USE_BLAS */
/**********/

/*** ARPACK ***/
/* prototype for the fortran function */
extern void ardrive_(int *N, double *AA,
                int *nev, int *ncv, int *lworkl, double *resid,
                double *v, double *workd, double *workl, int *info, int *ido,
                double *select_, double *d, int *ierr,
                int *maxit);

/* N -size of matrix
 * nev-no. of eigenvalues to find
 * Evals - array to store found eigenvalues, size nev x 1
 * Evecs - matrix to store found eigenvectos, size N x nev
 * A - NxN matrix, column major
 * maxiter: max no of iterations
 */
void
my_dsaupd(int N, int nev, double *Evals, double *Evecs, double *A, int maxiter)
{
  int ido, ncv;
	int i,j;

	double *resid, *v;
  double *workd, *workl;
  int lworkl, info, ierr;
  double *d, *selection;

	ncv=4*nev; /* nev-no. of eigenmodes found, so the size of 
										Krylov supspace is 4*nev */
	if(ncv>N) ncv=N; /* cannot exceed N */

	info=0; /* convergence information, initially set to 0
             to start with a random vector */

	if((selection=(double*)calloc((size_t)ncv,sizeof(double)))==0) {
		 fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
		 exit(1);
	}

	lworkl=ncv*ncv+8*ncv; /* work length */


	/* work area */
	if((workd=(double*)calloc((size_t)N*3,sizeof(double)))==0) {
		 fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
		 exit(1);
	} 
 if((workl=(double*)calloc((size_t)lworkl,sizeof(double)))==0) {
		 fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
		 exit(1);
	} 
	/* eigenvectors */
	if((d=(double*)calloc( (size_t)(ncv*2),sizeof(double)))==0) {
		 fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
		 exit(1);
	} 
	if((v=(double*)calloc((size_t)N*ncv, sizeof(double)))==0) {
		 fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
		 exit(1);
	} 
	if((resid=(double*)calloc( (size_t)N, sizeof(double)))==0) {
		 fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
		 exit(1);
	} 
  ido=0; /* initialize reverse comm parameter */

 printf("ARPACK: N=%d, nev=%d, ncv=%d lworkl=%d\n",N,nev,ncv,lworkl);
  ardrive_(&N, A,
          &nev, &ncv, &lworkl, resid,
          v, workd, workl, &info, &ido,
          selection, d, &ierr,
          &maxiter);

 printf("info=%d, ido=%d, ierr=%d\n",info,ido,ierr);


	/* copy the solution back */
	for(i=0; i<nev; i++){
			Evals[i]=d[nev-1-i];
#ifdef DEBUG
			printf("%lf\n",d[nev-1-i]);
#endif
			for(j=0; j<N; j++) {
       /* reverse the eigenvectors, because we are sorting in descending order */
				Evecs[(nev-i-1)*N+j]=v[i*N+j];
#ifdef DEBUG
			  printf("%lf\n",v[i*N+j]);
#endif
			}
	}

#ifdef DEBUG
	for(i=0;i<ncv*2; i++)
		printf("%lf ",d[i]);
	printf("\n");
#endif
	free(selection);
	
  free(workd);
	free(resid);
	free(v);
	free(workl);
	free(d); 
}
/***************/

static int
select_file(const struct dirent *dent)
{
	if (!strcmp(dent->d_name,"..")) return 0;
  if (strlen(dent->d_name)>1)
   return 1;

	return 0;
}

/* Execute the command using this shell program.  */
#define SHELL "/bin/sh"

int
my_system (const char *command)
{
       int status;
       pid_t pid;

       pid = fork ();
       if (pid == 0)
         {
           /* This is the child process.  Execute the shell command. */
           execl (SHELL, SHELL, "-c", command, NULL);
           _exit (EXIT_FAILURE);
         }
       else if (pid < 0)
         /* The fork failed.  Report failure.  */
         status = -1;
       else
         /* This is the parent process.  Wait for the child to complete.  */
         if (waitpid (pid, &status, 0) != pid)
           status = -1;
       return status;
}

/* limits: given by lpix,hpix
 * correlated data: saved in r , size MN x MN, determined by pixels
 */
static int
correlate_fits_file(const char *filename, long int* lpix, long int *hpix, double *r, long int rl) {
  int status;
  fitsfile *fptr;
	double bscale, bzero;
	long int totalpix;
	double *myarr;
	long int increment[4]={1,1,1,1};
	int null_flag=0;

	long int M,N, i,j;

	status=0;
	fits_open_file(&fptr, filename, READONLY, &status); 

	/* turn off scaling so that we copy the pixel values */
	bscale=1.0; bzero=0.0;
	fits_set_bscale(fptr,  bscale, bzero, &status);

	M=hpix[0]-lpix[0]+1;
	N=hpix[1]-lpix[1]+1;

	/* calculate total number of pixels */
	totalpix=(hpix[3]-lpix[3]+1)*(hpix[2]-lpix[2]+1)*M*N;
	printf("reading %ld pixels\n",totalpix);
	if ((myarr=(double*)calloc((size_t)totalpix,sizeof(double)))==0) {
	  fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
	  return 1;
	}

	fits_read_subset(fptr, TDOUBLE, lpix, hpix, increment,
	   0, myarr, &null_flag, &status);
			

/*	for (i=0; i<M; i++) {
	  for(j=0; j<N; j++) {
			printf("%lf ",myarr[j*M+i]);
		}
		printf("\n");
	} */

	/* calculate correlation */
  for (j=0; j<M*N; j++) { /* column */
		for(i=0; i<=j; i++) { /* row - only upper triangle */
			//printf("%lf %lf %lf\n",(*r)[j*(i+1)/2+i],myarr[i],myarr[j]);
			r[j*(j+1)/2+i]+=myarr[i]*myarr[j];
			//r[j*M*N+i]+=myarr[i]*myarr[j];
		}
	}

	fits_close_file(fptr,&status);
	if (status)
		fits_report_error(stderr, status); 


	free(myarr);
  return 0;
}



/*
 * filelist: array of FITS files, size nfilesx1
 * MxN: defined by the size of lpix, hpix pixel positions
 * A - size MN x MN - column major matrix stored as a  vector : will store the
 *  eigenvectors
 * S - size MN singular values
 * if use_arpack=1, only howmany eigenvalues, eigenvectors will be calculated
 */
int 
correlate_fits_filelist(char **filelist, int nfiles, long int *lpix, long int *hpix, double *A, double *S, int use_arpack, int howmany) {

	 int cnt;

	 long int M,N, rlen;

	 double *WORK;
	 int IWORK,i,j;

	 int *LWORK;
	 double *VT;

   /* for arpack */
   double *r,*V;

   double *R;

	 M=hpix[0]-lpix[0]+1;
	 N=hpix[1]-lpix[1]+1;
	 /* allocate memory for correlated values */
	 /* matrix size: MN x MN , however, we only store the upper triangle*/
	 rlen=(M*N)*(M*N+1)/2;
	 if ((R=(double*)calloc((size_t)rlen,sizeof(double)))==0) {
		 fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
		 return 1;
   } 
	for (cnt = 0; cnt < nfiles; ++cnt) {
			 /* create full path */
			 printf("processing file %s\n",filelist[cnt]);
       correlate_fits_file(filelist[cnt], lpix, hpix, R, rlen);
	}

	printf("M=%ld, N=%ld, R=%ld\n",M,N,rlen);

	/* copy triangle A to full A*/
	/* normalize correlations */
  for (j=0; j<M*N; j++) { /* column */
		for(i=0; i<j; i++) { /* row - only upper triangle */
			(A)[j*N*M+i]= (A)[i*N*M+j]=R[j*(j+1)/2+i]/(double)nfiles;
			//(A)[i*N*M+j]=(A)[j*N*M+i]=(A)[j*N*M+i]/(double)nfiles;
		}
		/* i = j, diagonal */
    (A)[j*(N*M+1)]=R[j*(j+1)/2+i]/(double)nfiles;
	  //(A)[j*(N*M+1)]/=(double)nfiles;
	}
 free(R);

 if (!use_arpack) {
	/* setup LAPACK */
	if ((WORK=(double*)calloc((size_t)10,sizeof(double)))==0) {
		 fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
		 return 1;
  }
  if ((VT=(double*)calloc((size_t)M*N*M*N,sizeof(double)))==0) {
		 fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
		 return 1;
   }
   if ((LWORK=(int*)calloc((size_t)8*M*N,sizeof(int)))==0) {
		 fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
		 return 1;
   }
	/* divide and conquor */
  cnt=dgesdd('O', M*N, M*N, A, M*N, S, 0,
				M*N, VT, M*N, WORK, -1, LWORK);
	IWORK=(int)WORK[0];
	printf("info= %d, work size =%d\n",cnt,IWORK);
  if ((WORK=(double*)realloc(WORK,(size_t)IWORK*sizeof(double)))==0) {
		 fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
		 return 1;
  }
  cnt=dgesdd('O', M*N, M*N, A, M*N, S, 0,
				M*N, VT, M*N, WORK, IWORK, LWORK);
	free(VT);
	free(LWORK);
#ifdef DEBUG
	/* print first 10 singular values */
	for (cnt = 0; cnt < 10; cnt++)
		printf("%d %lf\n",cnt,(S)[cnt]);
#endif
	free(WORK);
  } else {
   /* use arpack */
   /* eigenvalues */
  if ((r=(double *)calloc((size_t)howmany,sizeof(double)))==0) {
    fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
    exit(1);
  }
  /* eigenvectors */
  if ((V=(double *)calloc((size_t)N*M*howmany,sizeof(double)))==0) {
    fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
    exit(1);
  }

  /* call ARPACK */
  my_dsaupd(M*N,howmany,r,V, A, 300);

  /* copy back */
  for (i=0; i<M*N*howmany; i++) {
    A[i]=V[i];
  }
  for(j=0; j<howmany; j++) {
		S[j]=r[j];
  }

  free(r);
  free(V);

  }


	return 0;
}


/* write array to fits file
 */
int 
write_fits_file(const char *filename, long int* lpix, long int *hpix,double *A, long int M, long int N, long int nmodes, int dowhat, int unique_mode) {
  int status=0;
  long int increment[4]={1,1,1,1};
	int null_flag=0;
  double *myvec,*b;

  double alpha;
  long int i,totalpix;

#ifndef USE_BLAS
  long int j;
#endif

   fitsfile *fptr;
   double bscale, bzero;

	 fits_open_file(&fptr, filename, READWRITE, &status); 
	 /* turn off scaling so that we copy the pixel values */
	 bscale=1.0; bzero=0.0;


   /* allocate memory for reading */
	 totalpix=(hpix[0]-lpix[0]+1)*(hpix[1]-lpix[1]+1);

  	printf("reading %ld pixels\n",totalpix);
  	if ((myvec=(double*)calloc((size_t)totalpix,sizeof(double)))==0) {
  	  fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
  	  return 1;
	  }

	  fits_set_bscale(fptr,  bscale, bzero, &status);
	  fits_read_subset(fptr, TDOUBLE, lpix, hpix, increment,
	   0, myvec, &null_flag, &status);
  	if ((b=(double*)calloc((size_t)totalpix,sizeof(double)))==0) {
  	  fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
  	  return 1;
	  }

    if (dowhat==SUBTRACT_MODES) {
     memcpy(b,myvec,(size_t)totalpix*sizeof(double));
    }
			
    if (dowhat==PRESERVE_UNIQUE_MODE) {
     /* we need only a single mode, assume this is present */
      i=unique_mode;
      alpha=0;
#ifdef USE_BLAS
      alpha=my_ddot((int)N*M, myvec, &A[i*N*M]);
#else
      for(j=0; j<totalpix; j++) alpha+=myvec[j]*A[i*N*M+j];
#endif
      printf("mode %ld alpha=%lf\n",i,alpha);
      /* use magnitude */
      //if(alpha>0) alpha=-alpha;
#ifdef USE_BLAS
      my_daxpy((int)N*M,&A[i*N*M],alpha,b);
#else
      for(j=0; j<totalpix; j++) b[j]+=alpha*A[i*N*M+j];
#endif
    } else {
    /* accumulate product */
    for (i=0; i<nmodes; i++) {
      alpha=0;
#ifdef USE_BLAS
      alpha=my_ddot((int)N*M, myvec, &A[i*N*M]);
#else
      for(j=0; j<totalpix; j++) alpha+=myvec[j]*A[i*N*M+j];
#endif
      printf("mode %ld alpha=%lf\n",i,alpha);
      if (dowhat==SUBTRACT_MODES) {
       alpha=-alpha;
      } else if (dowhat==PRESERVE_MODES) {
      }
#ifdef USE_BLAS
      my_daxpy((int)N*M,&A[i*N*M],alpha,b);
#else
       for(j=0; j<totalpix; j++) b[j]+=alpha*A[i*N*M+j];
#endif
    }
    }

    free(myvec);

	 fits_set_bscale(fptr,  bscale, bzero, &status);

   fits_write_subset(fptr, TDOUBLE, lpix, hpix, b, &status);
   fits_close_file(fptr, &status);      /* all done */
   free(b);
   if (status)
      fits_report_error(stderr, status);  /* print out error messages */
   return(status);
}


/* 
 * A: size MxN times MxN matrix
 * write first nfiles columns of A (size MxN)  to each fits file
 */
int
subtract_mode_filelist(char **filelist, int nfiles, long int *lpix, long int *hpix, double *A, long int M, long int N, long int nmodes, int dowhat, int unique_mode) {
  int cnt;
	for (cnt = 0; cnt < nfiles; ++cnt) {
			 /* create full path */
			 printf("writing to file %s\n",filelist[cnt]);
       write_fits_file(filelist[cnt], lpix, hpix, A, M, N, nmodes, dowhat, unique_mode);
	}

  return 0;
}

void
print_help(void) {
   fprintf(stderr,"Usage:\n");
   fprintf(stderr,"-d fits_dir (-s no.of.modes | -p no.of modes | -u mode.number) -g pixels -t stokes -a\n");
   fprintf(stderr,"fits_dir: directory where there are ONLY *.fits files (mandatory)\n");
   fprintf(stderr,"-s no.of.modes: these modes will be subtracted\n");
   fprintf(stderr,"-p no.of.modes: only these modes will be preserved\n");
   fprintf(stderr,"-u mode.number: only this mode will be preserved, can be 0,1,2,...\n");
   fprintf(stderr,"note that you can only use -s, -p or -u, not all together\n");
   fprintf(stderr,"-g pixels: divide image to [pixels x pixels] subimages, if g is 0, whole image is used\n");
   fprintf(stderr,"-t stokes: which stokes plane 0 (I) 1 (Q) 2 (U) 3 (V), default I\n");
   fprintf(stderr,"-a : use ARPACK, iterative, for large images\n");
}


/* for getopt() */
extern char *optarg;
extern int optind, opterr, optopt;

int main(int argc, char **argv) {

	 struct dirent **eps;
	 int n;
	 int cnt;
   int dowhat=-1;
   int unique_mode=0;
   int use_arpack=0;
   int stokes=0;

	 long int M,N;
   long int step=30;
	 long int lpix[4]={1,1,1,1};
	 long int hpix[4]={1,1,1,1};
   int status=0;
   int naxis;
   long int *dims;
   
#ifdef DEBUG
	 FILE *dbg;
#endif

	 double *A,*S;
	 int i,j;
   int c;


	 char *basename=0;
	 char **filelist=0;
   int nmodes=6; /* number of eigenmodes */

   while ((c=getopt(argc,argv,"d:s:p:u:g:t:a"))!=-1) {
     switch(c) {
       case 'd':
        if (optarg) {
         basename=(char*)calloc((size_t)strlen((char*)optarg)+1,sizeof(char));
         if ( basename== 0 ) {
          fprintf(stderr,"%s: %d: no free memory",__FILE__,__LINE__);
          exit(1);
         }
         strcpy(basename,(char*)optarg);
        }
       break;
       case 's':
        if (optarg) {
         nmodes=strtol(optarg,0,0);
         if(!nmodes) nmodes=2;
        }
       dowhat=SUBTRACT_MODES;
       break;
       case 't':
        if (optarg) {
         stokes=atoi(optarg);
         if(stokes>3 || stokes<0) {
		      fprintf(stderr,"%s: %d: invalid stokes parameter %d\n",__FILE__,__LINE__,stokes);
          stokes=0;
         }
        }
       break;
       case 'a':
        use_arpack=1;
       break;
       case 'p':
        if (optarg) {
         nmodes=strtol(optarg,0,0);
         if(!nmodes) nmodes=2;
        }
       dowhat=PRESERVE_MODES;
       break;
       case 'u':
        if (optarg) {
         unique_mode=strtol(optarg,0,0);
         if(unique_mode<0) unique_mode=0;
         nmodes=unique_mode+1;
        }
       dowhat=PRESERVE_UNIQUE_MODE;
       break;
       case 'g':
        if (optarg) {
         step=strtol(optarg,0,0);
         if(step<0) step=30;
        }
       break;
       default:
        print_help();
       break;
     }
   }

   if (!basename) {
     print_help();
     exit(1);
   }

	 n = scandir (basename, &eps, select_file, alphasort);
   printf("found %d files\n",n);
	 if (n<0) {
		 fprintf(stderr,"%s: %d: invalid directory %s\n",__FILE__,__LINE__,basename);
		 return 1;
	 }
   /*create full file list */
   if ((filelist=(char**)calloc((size_t)n,sizeof(char*)))==0) {
		  fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
		  return 1;
   }

	for (cnt = 0; cnt < n; ++cnt) {
			 /* create full path */
       if ((filelist[cnt]=(char*)malloc((strlen(basename)+strlen(eps[cnt]->d_name)+1)*sizeof(char)))==0) {
		     fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
		     return 1;
       }
			 strcpy(filelist[cnt],basename);
			 strcat(filelist[cnt],eps[cnt]->d_name);
			 printf("adding file %d %s\n",cnt,filelist[cnt]);
  }

  free(basename);
	/* free memory  before calling LAPACK */
	for (cnt = 0; cnt < n; ++cnt)
			free(eps[cnt]);
	free(eps);

  /* read first file for size */
  fitsfile *fptr;
	printf("opening file %s\n",filelist[0]);
  fits_open_file(&fptr, filelist[0] , READONLY, &status);
  if (status) {
      /* if cannot open files, this is critical, so exit */
      fits_report_error(stderr, status); 
		  fprintf(stderr,"%s: %d: cannot open file %s\n",__FILE__,__LINE__,filelist[0]);
      exit(1);
  }
  fits_get_img_dim(fptr, &naxis, &status);
  printf("axes %d\n",naxis);
  if ((dims=(long int*)calloc((size_t)naxis,sizeof(long int)))==0) {
		  fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
		  return 1;
  }
  fits_get_img_size(fptr, naxis, dims, &status);
  fits_close_file(fptr,&status);
  for(i=0; i<naxis;i++)
   printf("axes %d %ld ",i,dims[i]);
  printf("\n");

  printf("stokes=%d\n",stokes);
  /* partition images for correlation */
  lpix[0]=lpix[1]=lpix[3]=1;
  hpix[3]=1;
  lpix[2]=hpix[2]=stokes+1;
  /* partition if step size is > 0 */
  if (step>0) {
  M=dims[0]/step;
  N=dims[1]/step;
  if (M*step<dims[0]) M+=1;
  if (N*step<dims[1]) N+=1;
  printf("partitions %ld x %ld\n",M,N);
  /* allocate memory */
	if ((A=(double*)calloc((size_t)step*step*step*step,sizeof(double)))==0) {
		 fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
		 return 1;
  }
	if ((S=(double*)calloc((size_t)step*step,sizeof(double)))==0) {
		 fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
		 return 1;
  }

  for (i=0; i<M; i++) { //M
   lpix[0]=i*step+1;
   for (j=0; j<N; j++) { //N
    lpix[1]=j*step+1;
    hpix[1]=lpix[1]+step-1;
    hpix[0]=lpix[0]+step-1;
    /* check for boundary */
    if (hpix[0]>dims[0]) hpix[0]=dims[0];
    if (hpix[1]>dims[1]) hpix[1]=dims[1];

    printf("partitions {%ld,%ld}x{%ld,%ld}\n",lpix[0],lpix[1],hpix[0],hpix[1]);
    /* perform correlation, and calculate SVD */
    correlate_fits_filelist(filelist,n,lpix,hpix,A,S,use_arpack,nmodes);
    /* subtract the first nmodes eigenmodes from each fits file */
    subtract_mode_filelist(filelist,n,lpix,hpix,A,hpix[0]-lpix[0]+1,hpix[1]-lpix[1]+1,nmodes,dowhat,unique_mode);
    /* reset memory */
    memset((void*)A,0,step*step*step*step);
    memset((void*)S,0,step*step);
   }
  }

   free(A);
   free(S);

  } else {
   /* read the full image */
   M=dims[0];
   N=dims[1];
   hpix[0]=lpix[0]+M-1;
   hpix[1]=lpix[1]+N-1;
   printf("partitions {%ld,%ld}x{%ld,%ld}\n",lpix[0],lpix[1],hpix[0],hpix[1]);

   /* allocate memory */
	 if ((A=(double*)calloc((size_t)M*N*M*N,sizeof(double)))==0) {
	 	 fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
		 return 1;
   }
	 if ((S=(double*)calloc((size_t)M*N,sizeof(double)))==0) {
		 fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
		 return 1;
   }
   correlate_fits_filelist(filelist,n,lpix,hpix,A,S,use_arpack,nmodes);
   subtract_mode_filelist(filelist,n,lpix,hpix,A,hpix[0]-lpix[0]+1,hpix[1]-lpix[1]+1,nmodes,dowhat,unique_mode);

   free(A);
   free(S);
  }


	for (cnt = 0; cnt < n; ++cnt)
			free(filelist[cnt]);
  free(filelist);

  free(dims);

  if (status)
      fits_report_error(stderr, status);  /* print out error messages */
	return 0;
}
